' Transpiled on 23-11-2022 11:03:02

'----------------------
'Raycast.bas by Martin Herhaus
'Raycast Engine MMBasic translated from
'https://youtu.be/gYRrGTC7GtA by 3DSage
'----------------------

' BEGIN:     #include "src/ctrl.ipp" -------------------------------------------
' Copyright (c) 2022 Thomas Hugo Williams
' License MIT <https://opensource.org/licenses/MIT>
'
' MMBasic Controller Library
'
' Preprocessor flag CMM2 defined
' Preprocessor flag CTRL_ONE_PLAYER defined
' Preprocessor flag CTRL_NO_SNES defined
' Preprocessor flag CTRL_USE_KEYDOWN defined

Const ctrl.VERSION = 905  ' 0.9.5

' Button values as returned by controller driver subroutines.
Const ctrl.R      = &h01
Const ctrl.START  = &h02
Const ctrl.HOME   = &h04
Const ctrl.SELECT = &h08
Const ctrl.L      = &h10
Const ctrl.DOWN   = &h20
Const ctrl.RIGHT  = &h40
Const ctrl.UP     = &h80
Const ctrl.LEFT   = &h100
Const ctrl.ZR     = &h200
Const ctrl.X      = &h400
Const ctrl.A      = &h800
Const ctrl.Y      = &h1000
Const ctrl.B      = &h2000
Const ctrl.ZL     = &h4000

Const ctrl.OPEN  = -1
Const ctrl.CLOSE = -2
Const ctrl.SOFT_CLOSE = -3

' The NES standard specifies a 12 micro-second pulse, but all the controllers
' I've tested work with 1 micro-second, and possibly less.
Const ctrl.PULSE = 0.001 ' 1 micro-second

' When a key is down the corresponding byte of this 256-byte map is set,
' when the key is up then it is unset.
'
' Note that when using INKEY$ (as opposed to the CMM2 'KEYDOWN' function or
' the PicoMiteVGA 'ON PS2' command) to read the keyboard we cannot detect
' keyup events and instead automatically clear a byte after it is read.
Dim ctrl.key_map%(31 + Mm.Info(Option Base))

' Timer number configured for reading the KEYDOWN state on the CMM2.
Dim ctrl.tick_nbr%

' Initialises keyboard reading.
'
' @param  period%  CMM2 only - interval to read KEYDOWN state, default 40 ms.
' @param  nbr%     CMM2 only - timer nbr to read KEYDOWN state, default 4.
Sub ctrl.init_keys(period%, nbr%)
  ctrl.term_keys()
  ctrl.tick_nbr% = Choice(nbr% = 0, 4, nbr%)
  SetTick Choice(period% = 0, 40, period%), ctrl.on_tick(), ctrl.tick_nbr%
End Sub

' Note there is little point in calling KeyDown(0) to determine the number of
' keys that are down, hardware limitations mean it's unlikely ever to be > 4
' and if a given key isn't down it just returns 0 so we harmlessly set that
' byte in the key map.
Sub ctrl.on_tick()
  Memory Set Peek(VarAddr ctrl.key_map%()), 0, 256
  Poke Var ctrl.key_map%(), KeyDown(1), 1
  Poke Var ctrl.key_map%(), KeyDown(2), 1
  Poke Var ctrl.key_map%(), KeyDown(3), 1
  Poke Var ctrl.key_map%(), KeyDown(4), 1
End Sub

' Terminates keyboard reading.
Sub ctrl.term_keys()
  If ctrl.tick_nbr% <> 0 Then SetTick 0, 0, ctrl.tick_nbr%
  Memory Set Peek(VarAddr ctrl.key_map%()), 0, 256
  Do While Inkey$ <> "" : Loop
End Sub

Function ctrl.keydown%(i%)
  ctrl.keydown% = Peek(Var ctrl.key_map%(), i%)
End Function

Function ctrl.poll_multiple$(ctrls$(), mask%, duration%)
  Local expires% = Choice(duration%, Timer + duration%, &h7FFFFFFFFFFFFFFF), i%
  Do
    For i% = Bound(ctrls$(), 0) To Bound(ctrls$(), 1)
      If ctrl.poll_single%(ctrls$(i%), mask%) Then
        ctrl.poll_multiple$ = ctrls$(i%)
        Exit Do
      EndIf
    Next
  Loop While Timer < expires%
End Function

' Opens, polls (for a maximum of 5ms) and closes a controller.
'
' @param  ctrl$  controller driver function.
' @param  mask%  bit mask to match against.
' @return        1 if any of the bits in the mask match what is read from the
'                controller, otherwise 0.
Function ctrl.poll_single%(ctrl$, mask%)
  On Error Ignore
  Call ctrl$, ctrl.OPEN
  If Mm.ErrNo = 0 Then
    Local key%, t% = Timer + 5
    Do
      Call ctrl$, key%
      If key% And mask% Then
        ctrl.poll_single% = 1
        ' Wait for user to release key.
        Do While key% : Pause 5 : Call ctrl$, key% : Loop
        Exit Do
      EndIf
    Loop While Timer < t%
    Call ctrl$, ctrl.SOFT_CLOSE
  EndIf
  On Error Abort
End Function

' Gets a string representation of bits read from a controller driver.
'
' @param  x%  bits returned by driver.
' @return     string representation.
Function ctrl.bits_to_string$(x%)
  Static BUTTONS$(14) = ("R","Start","Home","Select","L","Down","Right","Up","Left","ZR","X","A","Y","B","ZL")

  If x% = 0 Then
    ctrl.bits_to_string$ = "No buttons down"
    Exit Function
  EndIf

  ctrl.bits_to_string$ = Str$(x%) + " = "
  Local count%, i%, s$
  For i% = 0 To Bound(BUTTONS$(), 1)
    If x% And 2^i% Then
      s$ = BUTTONS$(i%)
      If count% > 0 Then Cat ctrl.bits_to_string$, ", "
      Cat ctrl.bits_to_string$, s$
      Inc count%
    EndIf
  Next
End Function

' Reads the keyboard as if it were a controller.
'
' Note that the PicoMite has no KEYDOWN function so we are limited to
' reading a single keypress from the input buffer and cannot handle multiple
' simultaneous keys or properly handle a key being pressed and not released.
Sub keys_cursor(x%)
  If x% < 0 Then Exit Sub
  x% =    ctrl.keydown%(32)  * ctrl.A
  Inc x%, ctrl.keydown%(128) * ctrl.UP
  Inc x%, ctrl.keydown%(129) * ctrl.DOWN
  Inc x%, ctrl.keydown%(130) * ctrl.LEFT
  Inc x%, ctrl.keydown%(131) * ctrl.RIGHT
End Sub

' Atari joystick port on CMM2 Deluxe G2.
Sub atari_dx(x%)
  Select Case x%
    Case Is >= 0
      x% =    Not Pin(32) * ctrl.A
      Inc x%, Not Pin(33) * ctrl.B
      Inc x%, Not Pin(35) * ctrl.UP
      Inc x%, Not Pin(36) * ctrl.DOWN
      Inc x%, Not Pin(38) * ctrl.LEFT
      Inc x%, Not Pin(40) * ctrl.RIGHT
      Exit Sub
    Case ctrl.OPEN
      SetPin 32, Din, PullUp : SetPin 33, Din, PullUp : SetPin 35, Din, PullUp
      SetPin 36, Din, PullUp : SetPin 38, Din, PullUp : SetPin 40, Din, PullUp
    Case ctrl.CLOSE, ctrl.SOFT_CLOSE
      SetPin 32, Off : SetPin 33, Off : SetPin 35, Off
      SetPin 36, Off : SetPin 38, Off : SetPin 40, Off
  End Select
End Sub

' NES gamepad attached USING ADAPTER to Atari joystick port on CMM2 Deluxe G2.
'
' IMPORTANT! the adapter is required to swap the Male DB9 (CMM2) +5V supply on
' Pin 7 to Pin 6 on the Female DB9 (Gamepad).
'
'   Pin 38: Latch, Pin 40: Clock, Pin 36: Data
Sub nes_dx(x%)
  Select Case x%
    Case Is >= 0
      Pulse 38, ctrl.PULSE
      x% =    Not Pin(36) * ctrl.A      : Pulse 40, ctrl.PULSE
      Inc x%, Not Pin(36) * ctrl.B      : Pulse 40, ctrl.PULSE
      Inc x%, Not Pin(36) * ctrl.SELECT : Pulse 40, ctrl.PULSE
      Inc x%, Not Pin(36) * ctrl.START  : Pulse 40, ctrl.PULSE
      Inc x%, Not Pin(36) * ctrl.UP     : Pulse 40, ctrl.PULSE
      Inc x%, Not Pin(36) * ctrl.DOWN   : Pulse 40, ctrl.PULSE
      Inc x%, Not Pin(36) * ctrl.LEFT   : Pulse 40, ctrl.PULSE
      Inc x%, Not Pin(36) * ctrl.RIGHT  : Pulse 40, ctrl.PULSE
      Exit Sub
    Case ctrl.OPEN
      SetPin 36, Din : SetPin 38, Dout : SetPin 40, Dout
      Pin(38) = 0 : Pin(40) = 0
      nes_dx(0) ' Discard the first reading.
    Case ctrl.CLOSE, ctrl.SOFT_CLOSE
      SetPin 36, Off : SetPin 38, Off : SetPin 40, Off
  End Select
End Sub

Sub wii_internal(i2c%, x%, type%)
  Static is_open%(3)
  If is_open%(i2c%) = -1 Then Error "I2C" + Str$(i2c%) + " failed to open" : Exit Sub

  If x% >= 0 Then
    Select Case is_open%(i2c%)
      Case &hA4200101
        x% = Classic(B, i2c%)
'        Inc x%, (Classic(LY, i2c%) > 200) * ctrl.UP
'        Inc x%, (Classic(LY, i2c%) < 60)  * ctrl.DOWN
'        Inc x%, (Classic(LX, i2c%) < 60)  * ctrl.LEFT
'        Inc x%, (Classic(LX, i2c%) > 200) * ctrl.RIGHT
        If x% = &h7FFF Then x% = 0 ' Ignore this glitch.
      Case &hA4200000
        x% =    Nunchuk(Z,  i2c%) * ctrl.A
        Inc x%, Nunchuk(C,  i2c%) * ctrl.B
        Inc x%, (Nunchuk(JY, i2c%) > 200) * ctrl.UP
        Inc x%, (Nunchuk(JY, i2c%) < 60)  * ctrl.DOWN
        Inc x%, (Nunchuk(JX, i2c%) < 60)  * ctrl.LEFT
        Inc x%, (Nunchuk(JX, i2c%) > 200) * ctrl.RIGHT
    End Select
    Exit Sub
  EndIf

  Select Case x%
    Case ctrl.SOFT_CLOSE
      ' Do nothing
      Exit Sub

    Case ctrl.CLOSE
      Select Case is_open%(i2c%)
        Case &hA4200000
          Controller Nunchuk Close i2c%
        Case &hA4200101
          Controller Classic Close i2c%
      End Select
      is_open%(i2c%) = 0
      Exit Sub

    Case ctrl.OPEN
      If is_open%(i2c%) Then Exit Sub
      is_open%(i2c%) = -1
      Controller Nunchuk Open i2c%
      If Mm.ErrNo Then
        If InStr(Mm.ErrMsg$, "already OPEN") Then
          Error "I2C" + Str$(i2c%) + " already open"
        Else
          Error "I2C" + Str$(i2c%) + " not connected"
        EndIf
        Exit Sub
      EndIf
      is_open%(i2c%) = Nunchuk(T, i2c%)
      Select Case is_open%(i2c%)
        Case &hA4200000
          If Not(type% And &h01) Then
            Controller Nunchuk Close i2c%
            is_open%(i2c%) = -1
            Error "Nunchuck controller on I2C" + Str$(i2c%) + " not supported"
          EndIf
        Case &hA4200101
          Controller Nunchuk Close i2c%
          If Not(type% And &h10) Then
            is_open%(i2c%) = -1
            Error "Classic controller on I2C" + Str$(i2c%) + " not supported"
          Else
            Controller Classic Open i2c%
          EndIf
        Case Else
          Controller Nunchuck Close i2c%
          is_open%(i2c%) = -1
          Error "Unrecognised controller on I2C" + Str$(i2c%)
      End Select
  End Select
End Sub

' Wii Nunchuk OR Classic gamepad on I2C1.
Sub wii_any_1(x%)
  wii_internal(1, x%, &h11)
End Sub

' Wii Nunchuk OR Classic gamepad on I2C2.
Sub wii_any_2(x%)
  wii_internal(2, x%, &h11)
End Sub

' Wii Nunchuk OR Classic gamepad on I2C3.
Sub wii_any_3(x%)
  wii_internal(3, x%, &h11)
End Sub

' Wii Classic gamepad on I2C1.
Sub wii_classic_1(x%)
  wii_internal(1, x%, &h10)
End Sub

' Wii Classic gamepad on I2C2.
Sub wii_classic_2(x%)
  wii_internal(2, x%, &h10)
End Sub

' Wii Classic gamepad on I2C3.
Sub wii_classic_3(x%)
  wii_internal(3, x%, &h10)
End Sub

' Wii Nunchuk on I2C1.
Sub wii_nunchuk_1(x%)
  wii_internal(1, x%, &h01)
End Sub

' Wii Nunchuk on I2C2.
Sub wii_nunchuk_2(x%)
  wii_internal(2, x%, &h01)
End Sub

' Wii Nunchuk on I2C3.
Sub wii_nunchuk_3(x%)
  wii_internal(3, x%, &h01)
End Sub

' END:       #Include "src/ctrl.ipp" -------------------------------------------
Dim float px,py,pdy,pdx,pa
Dim integer fps%,bg%,FC%,n%,mapX,mapY,mapS,HV,HH,vmul,hnul
mapS=64:mapy=24:mapx=24
Dim Map%(mapy*mapx)
P2 =Pi/2
P3=3*Pi/2
P2I=2*PI
DR=0.0174533 'one degree in radians

'Init
Select Case Mm.Device$
Case "PicoMiteVGA"
 Mode 2
 Font 1
 FrameBuffer Create
 FrameBuffer Write F
 Dim DRIVERS$(2) = ("atari_a", "nes_a", "keys_wasd")
 Const CONTROLLER_TEXT$ = "       Press START, FIRE or SPACE       "
Case "Colour Maximite 2", "Colour Maximite 2 G2", "MMBasic for Windows"
 Mode 1,8 '800x600
 Font 1
 Page Write 1
 If Mm.Device$ = "MMBasic for Windows" Then
   Dim DRIVERS$(1) = ("keys_wasd", "keys_wasd")
   Const CONTROLLER_TEXT$ = "          Press SPACE to start          "
 Else
   Dim DRIVERS$(3) = ("atari_dx", "nes_dx", "wii_any_3", "keys_wasd")
   Const CONTROLLER_TEXT$ = "       Press START, FIRE or SPACE       "
 EndIf
Case Else
 Error "Unsupported device: " + Mm.Device$
End Select
vmul=MM.VRes/512
hmul=MM.HRes/1024
hv=MM.VRes/2
hh=MM.HRes/2

Restore MapData:For n%=0 To mapx*mapy-1:Read Map%(n%):Next n%
bg%= RGB(0,64,0)
CLS bg%
px=300:py=300
pdx=Cos(pa)*5:pdy=Sin(pa)*5
tm%=Timer:fps%=0

drawRays3D

Text hh, hv + 20, CONTROLLER_TEXT$, "CM", 1, 1, Rgb(Black), Rgb(White)
If Mm.Device$ = "PicoMiteVGA" Then FrameBuffer Copy F, N, B Else Page Copy 1 To 0
ctrl.init_keys()
Dim driver$ = ctrl.poll_multiple$(DRIVERS$(), ctrl.A Or ctrl.B Or ctrl.START)
Call driver$, ctrl.OPEN
Dim key% = 1
Do While key% <> 0 : Call driver$, key% : Loop
Text hh, hv + 20, Space$(Len(CONTROLLER_TEXT$)), "CM", 1, 1, bg%, bg%

'main
Do
' display
drawRays3D
buttons
 If Mm.Device$ = "PicoMiteVGA" Then
   FrameBuffer Copy F, N
Else
   Page Copy 1 To 0
 EndIf
inc fps%:
if fps%=500 then
tm%=timer-tm%

  text hh+20,hv+20,str$(500/(tm%/1000),4,2)+"   ":fps%=0:tm%=Timer
end if
Loop

Function dist(ax,ay,bx,by,ang)
  dist=Sqr((bx-ax)*(bx-ax)+(by-ay)*(by-ay))
End Function

Sub drawRays3D
 Local integer r,mx,my,mp,dof,xstp,wcl
 Local floadt rx,ry,ra,xo,yo,dis1
 ra=pa-DR*30:Inc ra,P2i*(ra<0):Inc ra,-P2I*(ra>P2I)
 Box 0,0,hh<<1,hv,,RGB(0,255,255),RGB(0,255,255)
 Box 0,hv/2,hh<<1,hv/2,,RGB(255,64,0),RGB(255,64,0)
 'Cast 120 Rays
 For r=0 To 119
     'check Horizontal Lines
     dof=0
     distH=1000000:hx=px:hy=py
     atan=-1/Tan(ra)
     If ra>Pi Then
       ry=((Int(py>>6))<<6)-0.0001:rx=(py-ry)*atan+px ' Looking up
       yo=-64:xo=-yo*atan
     else
       ry=((Int(py>>6))<<6)+64:rx=(py-ry)*atan+px ' Looking down
       yo=64:xo=-yo*atan
     EndIf
     If not ra Or ra=Pi Then rx=px:ry=py:dof=mapx 'looking straight left or right
     Do While dof<mapx
       mx=Int(rx)>>6:my=Int(ry)>>6:mp=my*mapx+mx
         If mp>0 And mp<mapX*mapY Then If map%(mp)<>0 Then
           hx=rx:hy=ry:distH=dist(px,py,hx,hy,ra):dof=mapx
         Else
           rx=rx+Xo:ry=ry+Yo:Inc dof
        End If
    Loop
   'check vertical Lines
   dof=0
    distV=1000000:vx=px:vy=py
   ntan=-Tan(ra)
   If ra>P2 And ra< P3 Then
     rx=((Int(px>>6))<<6)-0.0001:ry=(px-rx)*ntan+py ' Looking left
     xo=-64:yo=-xo*ntan
   else
     rx=((Int(px>>6))<<6)+64:ry=(px-rx)*ntan+py ' Looking right
     xo=64:yo=-xo*ntan
   End If
    If not ra Or ra=Pi Then rx=px:ry=py:dof=mapy 'looking straight up or down
    Do While dof<mapy
      mx=Int(rx)>>6:my=Int(ry)>>6:mp=my*mapx+mx
      If mp<(mapX*mapY)And mp>-1 Then If map%(mp)<>0 Then
        vx=rx:vy=ry:distV=dist(px,py,vx,vy,ra):dof=mapy
      Else
        rx=rx+Xo:ry=ry+Yo:Inc dof
     End If
    Loop
    If distH<distV Then
      rx=hx:ry=hy:dis1=DistH:wcl=RGB(0,64,255)
    else
      rx=vx:ry=vy:dis1=DistV:wcl=RGB(0,128,255)
    end if
   '
   '--- Draw 3D Walls
   '
    ca=pa-ra:Inc ca,P2I*(ca<0):Inc ca,-P2I*(ca>P2I)
    dis1=dis1*Cos(ca)
    lineH=int((mapS*hh)/dis1):lineH=Min(lineH,hv)
    lineO=int(lineh>>1)
    Box xstp,(hv>>1)-LineO,1+hh/60,Lineh,,wcl,wcl
    Inc xstp,hh/60
    Inc ra,dr/2:Inc ra,P2I*(Ra<0):Inc ra,-P2I*(ra>P2I)
Next
End Sub

Sub display
drawmap2d
drawPlayer
End Sub

Sub drawPlayer
'unused
Local xo,yo
xo=px*hmul:yo=py*vmul
fc%=RGB(yellow)
Box xo-1,yo-1,3,3,,fc%
Line xo,yo,xo+pdx*5*hmul,yo+pdy*5*vmul,,fc%
End Sub


Sub buttons
  Local k%
  Call driver$, k%
  If Not k% Then Call "keys_cursor", k%
  Select Case k%
    Case ctrl.A, ctrl.B
      Save Image "cast.bmp"
    Case ctrl.UP
      Inc px,2*pdx
      Inc py,2*pdy
    Case ctrl.DOWN
      Inc px,2*-pdx
      Inc py,2*-pdy
    Case ctrl.LEFT
      Inc pa,-0.1
      If pa<0 Then Inc pa,P2I
      pdx=Cos(pa)*5:pdy=Sin(pa)*5
    Case ctrl.RIGHT
      Inc pa, 0.1
      If pa>P2I Then Inc pa,-P2I
      pdx=Cos(pa)*5:pdy=Sin(pa)*5
  End Select
End Sub

Sub drawMap2D
'unused
Local integer x,y,xo,yo
YO=2*HV/mapy:xo=HH/mapx
xo=xo*hmul:yo=yo*vmul
For y=0 To mapy-1:For x=0 To mapx-1
   Box x*xo,y*yo,xo,yo,,0,RGB(white)*(map%(y*mapX+x)<>0)
 Next : Next
End Sub

Sub keys_wasd(x%)
  If x% < 0 Then Exit Sub
  x% = ctrl.keydown%(32) * ctrl.A
  Inc x%, ctrl.keydown%(Asc("w")) * ctrl.UP
  Inc x%, ctrl.keydown%(Asc("s")) * ctrl.DOWN
  Inc x%, ctrl.keydown%(Asc("a")) * ctrl.LEFT
  Inc x%, ctrl.keydown%(Asc("d")) * ctrl.RIGHT
End Sub

MapData:
Data 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
Data 1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,1
Data 1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1
Data 1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1
Data 1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,1
Data 1,0,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1
Data 1,0,1,0,0,0,0,1,0,1,0,1,0,1,0,1,1,0,0,0,1,1,1,1
Data 1,0,1,0,0,0,0,1,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,8
Data 1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1
Data 1,0,1,0,0,0,0,1,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,1
Data 1,0,0,0,0,0,0,1,0,0,0,0,0,0,0,1,1,0,0,0,1,1,1,1
Data 1,0,0,0,0,0,0,1,1,1,1,0,1,1,1,1,1,1,1,1,1,1,1,1
Data 1,1,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,1,1,1,1,1
Data 1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1
Data 1,1,1,1,1,1,0,1,1,1,1,0,1,1,1,1,1,1,1,1,1,1,1,1
Data 1,1,1,1,1,1,0,1,1,1,1,0,1,1,1,1,1,1,1,1,1,1,1,1
Data 1,0,0,0,0,0,0,0,0,1,1,0,1,1,0,0,0,0,0,1,0,0,0,1
Data 1,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,1,0,0,1,0,0,0,1
Data 1,0,0,0,0,0,0,0,0,1,1,0,1,1,0,0,0,0,0,1,1,0,1,1
Data 1,0,1,0,1,0,0,0,0,1,1,0,0,0,0,0,1,0,0,0,0,0,0,1
Data 1,0,0,1,0,0,0,0,0,1,1,0,1,1,0,0,0,0,0,1,1,0,1,1
Data 1,0,1,0,1,0,0,0,0,1,1,0,1,1,0,0,1,0,0,1,0,0,0,1
Data 1,0,0,0,0,0,0,0,0,1,1,0,1,1,0,0,0,0,0,1,0,0,0,1
Data 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
